import numpy as np
import pytest
import qiskit

import bluequbit


def test_basic_large_jobs():
    qc = qiskit.QuantumCircuit(24)
    qc.x(np.arange(24))

    dq_client1 = bluequbit.BQClient()
    result1 = dq_client1.run(qc, asynchronous=True)

    dq_client2 = bluequbit.BQClient()
    result2 = dq_client2.run(qc, asynchronous=True)

    res1 = dq_client1.wait(result1.job_id)
    assert res1.run_status == "COMPLETED"
    assert res1.error_message is None
    assert res1.job_id == result1.job_id
    print("Job1", res1)

    res2 = dq_client2.wait(result2.job_id)
    assert res2.run_status == "COMPLETED"
    assert res2.error_message is None
    assert res2.job_id == result2.job_id
    print("Job2", res2)
    print("Downloading job 2")
    with pytest.raises(bluequbit.exceptions.BQStatevectorTooLargeError) as e:
        res2.get_statevector()
    assert (
        e.value.message
        == f"Statevector is too large for {res2.num_qubits} qubits. Use .get_counts() instead."
    )


def test_very_large_job():
    dq = bluequbit.BQClient()

    qc_qiskit = qiskit.QuantumCircuit(40)
    for i in range(40):
        qc_qiskit.h(i)
    qc_qiskit.cx(0, 1)

    with pytest.raises(bluequbit.exceptions.BQJobNotCompleteError) as e_info:
        dq.run(qc_qiskit)
    assert e_info.value.run_status == "FAILED_VALIDATION"


def test_very_large_job_but_unused_qubits():
    # Almost the same as test_very_large_job, except that most qubits are
    # unused.
    dq = bluequbit.BQClient()

    qc_qiskit = qiskit.QuantumCircuit(40)
    qc_qiskit.h(0)
    qc_qiskit.cx(0, 1)
    r0 = dq.run(qc_qiskit)
    assert r0.run_status == "COMPLETED"



import os
import time

import cirq
import pytest
import qiskit

import bluequbit


@pytest.mark.skipif(
    "BLUEQUBIT_STRESS_TESTS" not in os.environ, reason="Requires env variable to be set"
)
def test_gpu_job():
    dq_client = bluequbit.BQClient()
    # time_now = "2022-10-19T13:05:17.917290Z"
    n = 10  # will get to GPU1
    qc_qiskit = qiskit.QuantumCircuit(n)
    for i in range(n):
        qc_qiskit.h(i)
    qc_qiskit.measure_all()

    job_result = dq_client.run(qc_qiskit, job_name="testing_gpu_qiskit", device="gpu")
    print(job_result)

    assert job_result.run_status == "COMPLETED"


@pytest.mark.skipif(
    "BLUEQUBIT_STRESS_TESTS" not in os.environ, reason="Requires env variable to be set"
)
def test_gpu_job_cirq_measure_gpu1():
    dq_client = bluequbit.BQClient()
    num_qubits = 32
    qc_cirq = cirq.Circuit()
    qubits = cirq.LineQubit.range(num_qubits)
    for i in range(num_qubits):
        qc_cirq.append(cirq.H(qubits[i]))
    job_result = dq_client.run(qc_cirq, job_name="testing_gpu_cirq_meas", device="gpu")
    assert job_result.run_status == "COMPLETED"
    print(job_result)

    assert job_result.run_time_ms < 13000  # tests 12651, 12708, 12683


@pytest.mark.skipif(
    "BLUEQUBIT_STRESS_TESTS" not in os.environ, reason="Requires env variable to be set"
)
def test_gpu_job_cirq_measure_gpu2():
    dq_client = bluequbit.BQClient()
    num_qubits = 33
    qc_cirq = cirq.Circuit()
    qubits = cirq.LineQubit.range(num_qubits)
    for i in range(num_qubits):
        qc_cirq.append(cirq.H(qubits[i]))
    job_result = dq_client.run(qc_cirq, job_name="testing_gpu_cirq_meas", device="gpu")
    assert job_result.run_status == "COMPLETED"
    print(job_result)

    assert job_result.run_time_ms < 13300  # tests 12994, 12976, 13071


@pytest.mark.skipif(
    "BLUEQUBIT_STRESS_TESTS" not in os.environ, reason="Requires env variable to be set"
)
def test_gpu_job_cirq_measure_gpu4():
    dq_client = bluequbit.BQClient()
    num_qubits = 34
    qc_cirq = cirq.Circuit()
    qubits = cirq.LineQubit.range(num_qubits)
    for i in range(num_qubits):
        qc_cirq.append(cirq.H(qubits[i]))
    job_result = dq_client.run(qc_cirq, job_name="testing_gpu_cirq_meas", device="gpu")
    assert job_result.run_status == "COMPLETED"
    print(job_result)

    assert job_result.run_time_ms < 14000  # tests: 13726, 13743, 13728


@pytest.mark.skipif(
    "BLUEQUBIT_STRESS_TESTS" not in os.environ, reason="Requires env variable to be set"
)
def test_gpu_job_cirq_measure_gpu8():
    dq_client = bluequbit.BQClient()
    n = 35
    qc = cirq.Circuit()
    qs = cirq.LineQubit.range(n)
    for i in range(n):
        qc.append(cirq.H(qs[i]))
    qc.append(cirq.measure(qs))
    job_result = dq_client.run(qc, job_name="testing_gpu_cirq", device="gpu")
    assert job_result.run_status == "COMPLETED"
    assert job_result.run_time_ms < 15500  # tests: 15128, 15131, 15275
    print(job_result)


@pytest.mark.skipif(
    "BLUEQUBIT_STRESS_TESTS" not in os.environ, reason="Requires env variable to be set"
)
def test_gpu_job_cirq_measure_gpu16():
    dq_client = bluequbit.BQClient()
    n = 36
    qc = cirq.Circuit()
    qs = cirq.LineQubit.range(n)
    for i in range(n):
        qc.append(cirq.H(qs[i]))
    qc.append(cirq.measure(qs))
    job_result = dq_client.run(qc, job_name="testing_gpu_cirq", device="gpu")
    assert job_result.run_status == "COMPLETED"
    print(job_result)


@pytest.mark.skipif(
    "BLUEQUBIT_STRESS_TESTS" not in os.environ, reason="Requires env variable to be set"
)
def test_gpu_job_cirq_8gpu_issue():
    dq_client = bluequbit.BQClient()
    n = 35
    qc = cirq.Circuit()
    qs = cirq.LineQubit.range(n)
    for i in range(n):
        qc.append(cirq.H(qs[i]))
    qc.append(cirq.measure(qs))
    r = dq_client.run(qc, job_name="testing_gpu_cirq", device="gpu", asynchronous=True)
    for _ in range(1200):
        res = dq_client.get(r.job_id)
        if res.run_status == "RUNNING":
            break
        time.sleep(1)
    else:
        raise AssertionError()

    time.sleep(10)
    r = dq_client.cancel(r.job_id)
    assert r.run_status == "CANCELED"



class BQBaseError(Exception):
    def __init__(self, status_code, message="Unknown Error Message"):
        self.status_code = status_code
        self.message = message
        super().__init__(f"Error code: {status_code}. {message}")


class BQUnauthorizedError(BQBaseError):
    def __init__(self):
        super().__init__(
            "BQ_ACCESS_UNAUTHORIZED",
            "Wrong API key. You can find your API key once you log in to https://app.bluequbit.io",
        )


class BQStatevectorTooLargeError(BQBaseError):
    def __init__(self, num_qubits):
        super().__init__(
            "BQ_JOB_STATEVECTOR_TOO_LARGE",
            f"Statevector is too large for {num_qubits} qubits. Use .get_counts() instead.",
        )


class BQJobNotCompleteError(BQBaseError):
    def __init__(self, job_id, run_status, error_message):
        super().__init__(
            "BQ_JOB_NOT_COMPLETE",
            f"Job {job_id} finished with status: {run_status}. {error_message}",
        )
        self.job_id = job_id
        self.run_status = run_status
        self.error_message = error_message


class BQJobCouldNotCancelError(BQBaseError):
    def __init__(self, job_id, run_status, error_message):
        super().__init__(
            "BQ_JOB_COULDNOT_CANCEL",
            f"Couldn't cancel job {job_id}. Finished status is {run_status}. {error_message}",
        )
        self.job_id = job_id
        self.run_status = run_status
        self.error_message = error_message



import datetime
import threading
import time

import pytest
import qiskit

import bluequbit

NUM_PER_THREAD = 3
NUM_THREADS = 3

qc_qiskit = qiskit.QuantumCircuit(2)
qc_qiskit.h(0)
qc_qiskit.x(1)

job_ids = [[] for _ in range(NUM_THREADS)]


def submit_task(dq_client, thread_no):
    for _ in range(NUM_PER_THREAD):
        job_ids[thread_no].append(dq_client.run(qc_qiskit, asynchronous=True).job_id)


@pytest.mark.timeout(120)
def test_stress_submit_jobs():
    dq_client = bluequbit.BQClient()
    time_now = datetime.datetime.now(datetime.timezone.utc)
    threads = []
    for i in range(NUM_THREADS):
        threads.append(
            threading.Thread(
                target=submit_task,
                args=(
                    dq_client,
                    i,
                ),
            )
        )
        threads[-1].start()

    for i in range(NUM_THREADS):
        threads[i].join()

    job_ids_flat = [item for sublist in job_ids for item in sublist]

    assert len(job_ids_flat) == NUM_THREADS * NUM_PER_THREAD

    completed = None
    for i in range(NUM_PER_THREAD * NUM_THREADS):
        completed = 0
        jobs = dq_client.search(
            created_later_than=time_now - datetime.timedelta(seconds=10)
        )
        for job_result in jobs:
            if (
                job_result.job_id in job_ids_flat
                and job_result.run_status == "COMPLETED"
                and job_result.error_message is None
            ):
                completed += 1
        print("Try", i, ". Num completed", completed)
        if completed == len(job_ids_flat):
            break
        time.sleep(5.0)

    assert completed == len(job_ids_flat)



import qiskit

import bluequbit


def test_ok():
    dq_client = bluequbit.BQClient()
    # time_now = "2022-10-19T13:05:17.917290Z"
    qc_qiskit = qiskit.QuantumCircuit(2)
    qc_qiskit.h(0)
    qc_qiskit.x(1)
    job_result = dq_client.run(qc_qiskit, job_name="testing")

    assert job_result.ok



import numpy as np
import pytest
import qiskit

import bluequbit
from bluequbit import job_metadata_constants


def make_simple_circuit():
    qc_qiskit = qiskit.QuantumCircuit(2)
    qc_qiskit.h(0)
    qc_qiskit.x(1)
    return qc_qiskit


def test_job_estimation():
    dq_client = bluequbit.BQClient()
    qc_qiskit = make_simple_circuit()
    results = dq_client.estimate(qc_qiskit)
    assert results.estimated_runtime == 100


def test_job_estimation_wrong_device_type():
    dq_client = bluequbit.BQClient()
    qc_qiskit = make_simple_circuit()
    # Should work
    dq_client.estimate(qc_qiskit, device="CPU")

    # Should fail
    with pytest.raises(Exception) as e_info:
        dq_client.estimate(qc_qiskit, device="supercpu")
    expected = "Invalid device type. Must be one of " + ", ".join(
        job_metadata_constants.DEVICE_TYPES
    )
    assert expected in str(e_info)


def test_job_estimation_large():
    dq_client = bluequbit.BQClient()
    qc = qiskit.QuantumCircuit(24)
    qc.x(np.arange(24))
    results = dq_client.estimate(qc)
    assert results.estimated_runtime == 187
    # assert results.device == "qsim_simulator"
    assert results.num_qubits == 24
    assert (
        results.warning_message
        == "This is just an estimate; the actual runtime may be less or more."
    )


def test_job_estimate_validation_too_many_qubits():
    # Too many qubits
    dq_client = bluequbit.BQClient()
    qc = qiskit.QuantumCircuit(38)
    qc.x(np.arange(38))
    result = dq_client.estimate(qc)
    expected_message = "Circuit contains more than 34 qubits, which is not supported for CPU backend. See https://app.bluequbit.io/docs for more details."
    assert expected_message in result.error_message

    with pytest.raises(bluequbit.exceptions.BQJobNotCompleteError) as e_info:
        dq_client.run(qc)
    assert e_info.value.run_status == "FAILED_VALIDATION"



import os

import numpy as np
import pytest
import qiskit

import bluequbit

qc_qiskit = qiskit.QuantumCircuit(2)
qc_qiskit.h(0)
qc_qiskit.x(1)


def test_basic_cancel_job():
    dq_client = bluequbit.BQClient()

    jobs = []
    num_jobs = 5
    for _ in range(num_jobs):
        jobs.append(dq_client.run(qc_qiskit, asynchronous=True))

    dq_client.cancel(jobs[-1].job_id)

    for job in jobs[:-1]:
        res = dq_client.wait(job.job_id)
        assert res.run_status == "COMPLETED"
        assert res.error_message is None

    with pytest.raises(bluequbit.exceptions.BQJobNotCompleteError) as e_info:
        dq_client.wait(jobs[-1].job_id)
    assert e_info.value.run_status == "CANCELED"


@pytest.mark.skipif(
    "BLUEQUBIT_STRESS_TESTS" not in os.environ, reason="Requires env variable to be set"
)
def test_long_running_cancel_job():
    dq_client = bluequbit.BQClient()
    num_qubits = 29
    qc_qiskit = qiskit.QuantumCircuit(num_qubits)
    qc_qiskit.x(np.arange(num_qubits))

    r = dq_client.run(qc_qiskit, asynchronous=True)
    for _ in range(1200):
        res = dq_client.get(r.job_id)
        if res.run_status == "RUNNING":
            break
    else:
        raise AssertionError()

    r = dq_client.cancel(r.job_id)
    assert r.run_status == "CANCELED"
    assert r.error_message == "Canceled by user"



import qiskit

import bluequbit


def test_metadata_json():
    dq_client = bluequbit.BQClient()
    qc_qiskit = qiskit.QuantumCircuit(3, 3)
    qc_qiskit.x(0)
    qc_qiskit.measure(0, 0)
    qc_qiskit.measure(1, 1)
    qc_qiskit.measure(2, 2)
    job_result = dq_client.run(qc_qiskit, job_name="testing")
    print(job_result.get_counts())

    assert job_result._metadata["has_measurements"] is True
    assert job_result._metadata["message"] is None
    assert job_result._metadata["top_128"] == {"001": 1.0}
    assert job_result._metadata["counts"] == {"001": 1.0}
    assert job_result._metadata["job_id"] == job_result.job_id



JOB_NO_RESULT_TERMINAL_STATES = {
    "FAILED_VALIDATION",
    "CANCELED",
    "TERMINATED",
    "NOT_ENOUGH_FUNDS",
}
JOB_RESULTS_READY_STATES = {"COMPLETED"}
JOB_TERMINAL_STATES = JOB_RESULTS_READY_STATES.union(JOB_NO_RESULT_TERMINAL_STATES)
JOB_NON_TERMINAL_STATES = {"PENDING", "QUEUED", "RUNNING"}
JOB_STATES = JOB_NON_TERMINAL_STATES.union(JOB_TERMINAL_STATES)

DEVICE_TYPES = {"cpu", "gpu", "quantum"}



import braket.circuits
import cirq
import pytest
import qiskit

from bluequbit import circuit_serialization


def test_cirq():
    qc_cirq = cirq.Circuit()
    qubits = cirq.LineQubit.range(2)
    qc_cirq.append(cirq.H(qubits[0]))
    qc_cirq.append(cirq.X(qubits[1]))

    encoded_cirq = circuit_serialization.encode_circuit(qc_cirq)
    decoded_cirq = circuit_serialization.decode_circuit(encoded_cirq)
    all_ops = list(decoded_cirq.all_operations())
    # Check that we recover the original circuit
    assert len(all_ops) == 2
    assert isinstance(all_ops[0].gate, cirq.ops.HPowGate), all_ops[0].gate
    assert isinstance(all_ops[1].gate, cirq.ops.XPowGate), all_ops[0].gate


def test_qiskit():
    qc_qiskit = qiskit.QuantumCircuit(2)
    qc_qiskit.h(0)
    qc_qiskit.x(1)

    encoded_qiskit = circuit_serialization.encode_circuit(qc_qiskit)
    decoded_qiskit = circuit_serialization.decode_circuit(encoded_qiskit)
    qc = decoded_qiskit
    assert qc.data[0].operation.name == "h"
    assert qc.data[1].operation.name == "x"


def test_braket():
    qc_braket = braket.circuits.Circuit().h(0).x(1)
    encoded_braket = circuit_serialization.encode_circuit(qc_braket)
    assert (
        encoded_braket["circuit"]
        == "OPENQASM 3.0;\nbit[2] b;\nqubit[2] q;\nh q[0];\nx q[1];\nb[0] = measure q[0];\nb[1] = measure q[1];"
    )
    with pytest.raises(Exception, match="Braket decoding is not yet supported."):
        circuit_serialization.decode_circuit(encoded_braket)


def test_unsupported():
    wrong = {}
    with pytest.raises(Exception) as e_info:
        circuit_serialization.encode_circuit(wrong)
        assert e_info == Exception("Unsupported circuit type", wrong)
    with pytest.raises(
        Exception, match="Wrong data structure. 'circuit_type' must be present."
    ):
        circuit_serialization.decode_circuit({})
    with pytest.raises(
        Exception, match="Wrong data structure. 'circuit' must be present."
    ):
        circuit_serialization.decode_circuit({"circuit_type": "Qiskit"})



import qiskit
import requests

import bluequbit


def test_get_job():
    dq_client = bluequbit.init()
    qc_qiskit = qiskit.QuantumCircuit(2)
    qc_qiskit.h(0)
    qc_qiskit.x(1)
    result = dq_client.run(qc_qiskit)

    assert result.num_qubits == 2
    print(result)

    assert result._results_path is not None

    response = requests.get(result._results_path + "statevector.txt", timeout=60.0)
    assert len(response.content) == 326

    assert result.get_statevector().shape == (4,)
    assert len(result.get_counts()) == 2



import os

import numpy as np
import pytest
import qiskit

import bluequbit


@pytest.mark.skipif(
    "BLUEQUBIT_STRESS_TESTS" not in os.environ, reason="Requires env variable to be set"
)
def test_long_running_cancel_job():
    dq_client = bluequbit.BQClient()
    num_qubits = 29
    qc_qiskit = qiskit.QuantumCircuit(num_qubits)
    qc_qiskit.x(np.arange(num_qubits))

    with pytest.raises(bluequbit.exceptions.BQJobNotCompleteError) as e:
        dq_client.run(qc_qiskit)
    assert e.run_status == "TERMINATED"



import datetime

import qiskit

import bluequbit


def test_basic_search_jobs():
    time_now = datetime.datetime.now(datetime.timezone.utc)
    dq_client = bluequbit.BQClient()
    # time_now = "2022-10-19T13:05:17.917290Z"
    qc_qiskit = qiskit.QuantumCircuit(2)
    qc_qiskit.h(0)
    qc_qiskit.x(1)
    job_result = dq_client.run(qc_qiskit, job_name="testing_basic_search_jobs")
    results = dq_client.search(
        run_status="COMPLETED",
        created_later_than=time_now - datetime.timedelta(seconds=10),
    )
    num_found = 0
    for result in results:
        if (
            job_result.job_id == result.job_id
            and job_result.job_name == "testing_basic_search_jobs"
        ):
            print(job_result)
            num_found += 1

    assert num_found == 1



import datetime
import os

import numpy as np
import pytest
import qiskit

import bluequbit


@pytest.mark.skipif(
    "BLUEQUBIT_STRESS_TESTS" not in os.environ, reason="Requires env variable to be set"
)
def test_quantum_job():
    dq_client = bluequbit.BQClient()
    # time_now = "2022-10-19T13:05:17.917290Z"
    qc_qiskit = qiskit.QuantumCircuit(2)
    qc_qiskit.h(0)
    qc_qiskit.x(1)
    qc_qiskit.measure_all()
    job_result = dq_client.run(qc_qiskit, job_name="testing_quantum", device="quantum")
    assert job_result.error_message is None
    assert job_result.run_status == "COMPLETED"
    assert job_result.device == "quantum"

    counts = job_result.get_counts()
    print(counts)
    total_counts = 0
    for _, v in counts.items():
        total_counts += v
    assert total_counts == 1000


def test_quantum_job_no_measurement():
    dq_client = bluequbit.BQClient()
    qc_qiskit = qiskit.QuantumCircuit(2)
    qc_qiskit.h(0)

    with pytest.raises(bluequbit.exceptions.BQJobNotCompleteError) as e_info:
        dq_client.run(qc_qiskit, job_name="testing_quantum", device="quantum")
    assert e_info.value.run_status == "FAILED_VALIDATION"


def check_aspen_time_availability():
    time_now_hour = datetime.datetime.now(datetime.timezone.utc).hour
    if (time_now_hour >= 4 and time_now_hour < 6) or (
        time_now_hour >= 15 and time_now_hour < 18
    ):
        return True
    return False


@pytest.mark.skipif(
    check_aspen_time_availability(), reason="Requires env variable to be set"
)
def test_quantum_job_when_aspen_not_available():
    dq_client = bluequbit.BQClient()
    # time_now = "2022-10-19T13:05:17.917290Z"
    num_qubits = 30
    qc_qiskit = qiskit.QuantumCircuit(num_qubits)
    qc_qiskit.x(np.arange(num_qubits))
    qc_qiskit.measure_all()
    with pytest.raises(bluequbit.exceptions.BQJobNotCompleteError) as e_info:
        dq_client.run(
            qc_qiskit, job_name="testing_quantum", device="quantum", asynchronous=True
        )
    assert e_info.value.run_status == "FAILED_VALIDATION"


@pytest.mark.skipif(
    "BLUEQUBIT_STRESS_TESTS" not in os.environ, reason="Requires env variable to be set"
)
def test_cancel_quantum_job():
    dq_client = bluequbit.BQClient()
    # time_now = "2022-10-19T13:05:17.917290Z"
    num_qubits = 30
    qc_qiskit = qiskit.QuantumCircuit(num_qubits)
    qc_qiskit.x(np.arange(num_qubits))
    qc_qiskit.measure_all()
    job_result = dq_client.run(
        qc_qiskit, job_name="testing_quantum", device="quantum", asynchronous=True
    )
    job_result = dq_client.cancel(job_result.job_id)

    assert job_result.run_status == "CANCELED"
